// This example implements a Matrix class containing a collection of column-vectors

#include "vector.hpp"


class Matrix
{
protected:
    Vector **storage; // we use this format to create a list of pointers to Vector objects
    unsigned int rows;
    unsigned int cols;

public:
    // constructor
    Matrix(const unsigned int r, const unsigned int c) : rows(r), cols(c)
    {
        storage=new Vector*[cols];  // we adopt the column-vectors representation of matrix

        for (unsigned int i=0; i<cols; i++)
            storage[i]=new Vector(rows);
    }

    // copy constructor
    Matrix(const Matrix &M)
    {
        rows=M.getRows();
        cols=M.getCols();

        storage=new Vector*[cols];

        for (unsigned int j=0; j<cols; j++)
        {    
            storage[j]=new Vector(rows);

            for (unsigned int i=0; i<rows; i++)
                getElem(i,j)=M.getElem(i,j);
        }
    }

    unsigned int getRows() const
    {
        return rows;
    }


    unsigned int getCols() const
    {
        return cols;
    }

    float &getElem(const unsigned int r, const unsigned int c)
    {
        return storage[c]->getElem(r);
    }

    float &getElem(const unsigned int r, const unsigned int c) const
    {
        return storage[c]->getElem(r);
    }

    Vector &getVector(const unsigned int c)
    {
        return *(storage[c]);
    }

    Vector &getVector(const unsigned int c) const
    {
        return *(storage[c]);
    }

    Matrix add(const Matrix &M) const
    {
        Matrix res(rows,cols);        

        if (M.getRows()==rows && M.getCols()==cols)
            for (unsigned int i=0; i<cols; i++)
                res.getVector(i).copy(getVector(i).add(M.getVector(i)));

        return res;
    }

    Matrix sub(const Matrix &M) const
    {
        Matrix res(rows,cols);

        if (M.getRows()==rows && M.getCols()==cols)
            for (unsigned int i=0; i<cols; i++)
                res.getVector(i).copy(getVector(i).sub(M.getVector(i)));

        return res;
    }

    Matrix transpose() const
    {
        Matrix res(cols,rows);

        for (unsigned int j=0; j<res.getCols(); j++)
            for (unsigned int i=0; i<res.getRows(); i++)
                res.getElem(i,j)=getElem(j,i);

        return res;
    }

    Vector vectorMul(const Vector &v) const
    {        
        Vector res(rows);

        if (v.getN()==cols)
        {
            Matrix trp=transpose();
                
            for (unsigned int i=0; i<res.getN(); i++)
                res.getElem(i)=trp.getVector(i).dot(v);
        }

        return res;
    }

    Matrix matrixPostMul(const Matrix &M) const
    {        
        Matrix res(rows,M.getCols());

        if (M.getRows()==cols)
            for (unsigned int i=0; i<res.getCols(); i++)
                res.getVector(i).copy(vectorMul(M.getVector(i)));

        return res;
    }

    void print() const
    {
        for (unsigned int i=0; i<rows; i++)
        {
            fprintf(stdout,"(%d,1)=>(%d,%d): ",i+1,i+1,cols);

            for (unsigned int j=0; j<cols; j++)
                fprintf(stdout,"%.3f ",getElem(i,j));

            fprintf(stdout,"\n");
        }
    }

    ~Matrix()
    {
        for (unsigned int i=0; i<cols; i++)
            delete storage[i];

        delete[] storage;
    }
};


int main()
{
    // definitions
    Vector v(2);
    Matrix M1(2,2);
    Matrix M2(2,2);

    // fill up variables
    v.getElem(0)=1.0F;
    v.getElem(1)=2.0F;

    M1.getElem(0,0)=1.0F; M1.getElem(0,1)=2.0F;
    M1.getElem(1,0)=3.0F; M1.getElem(1,1)=4.0F;

    M2.getElem(0,0)=5.0F; M2.getElem(0,1)=0.0F;
    M2.getElem(1,0)=0.0F; M2.getElem(1,1)=5.0F;

    fprintf(stdout,"vector components\n");
    v.print();

    fprintf(stdout,"first matrix elements\n");
    M1.print();

    fprintf(stdout,"second matrix elements\n");
    M2.print();

    // execute algebric operations on matrices
    Matrix M3=M1.add(M2);
    Vector w=M1.vectorMul(v);
    Matrix M4=M1.matrixPostMul(M2);

    // dump results
    fprintf(stdout,"matrix sum\n");
    M3.print();

    fprintf(stdout,"matrix-vector product\n");
    w.print();

    fprintf(stdout,"matrix-matrix product\n");
    M4.print();

    return 0;
}



